module TreeLab
    
    #    Version 1.2    (27/09/2012)
    #    Changelog
    #        1.    Created Binary Search Tree class
    #
    #    Version 1.1    (08/08/2012)
    #    Changelog
    #        1.    Changed node.leftChild to node.left, and node.rightChild to node.right
    #        2.    Merged preorder_traversal(*speed), inorder_traversal(*speed), postorder_traversal(*speed) into traversal(tree, option, *speed), where valid options are preorder, inorder, and postorder
    #        3.    Display text at end of traversal to indicate visit order: "Visited 9, 6, 5, 4"
    #        4.    Display order number during traversal
    #        5.    Tree animation now freezes at last frame
    #
    #    Version  1.0    (06/08/2012)
    #    Changelog
    #        1.    Created Binary Tree, Node classes to represent a binary tree and a tree node
    #        2.    Created view_tree to visualize the tree and preorder_traversal, inorder_traversal, postorder_traversal to animate the traversal of a tree
        
    @@tree = nil
    @@visited = nil
    @@sleep = 0.5
    
    def view_tree(tree)
        @@tree = tree
        raise "Value must be a BinaryTree!" if !(tree.class == TreeLab::BinaryTree || tree.class == TreeLab::BinarySearchTree)
        # Initializes the canvas and font type
        Canvas.init(30+(2**tree.heightOfTree)*30, 50*tree.heightOfTree+50, "TreeLab")
        Canvas::Font.new('bucketfont', :family => 'Helvetica', :size => 10)

        # calls recursive method to draw the tree using post-order traversal
        if tree.root == nil
            return true
        end
        preOrder(tree.root, ((2**tree.heightOfTree)*30)/2, 30, ((2**tree.heightOfTree)*30)/2)
        return true
    end
    
    def tree
        @@tree
    end
    
    def traversal(tree, option, *speed)
        view_tree(tree)
        if (!speed[0])
            @@sleep = 0.5
        else
            @@sleep = speed[0]
        end
        
        Canvas::Font.new('bucketfont', :family => 'Helvetica', :size => 10)

        @@visited = Array.new
        
        root = @@tree.root
        sleep(1)
        case option
        when :preorder
            preorder_traverse(root)
        when :inorder
            inorder_traverse(root)
        when :postorder
            postorder_traverse(root)
        else
            return "Invalid option. Valid options include :preorder, :inorder, :postorder"
        end
        sleep(1)
        
        text = "Visited: "
        
        @@visited.each_with_index {
            |val, index| 
            text = text + val.to_s
            if index != @@visited.length-1
                text = text + ", "
            end
        }
        return text
    end
    
    # Ruby Reference-based implementation of a Binary Tree
    class BinaryTree
        
        # root node
        attr_reader :root
        
        # class constructor
        def initialize()
        end
        
        def setRoot(node)
            raise "Value must be a Node!" if node.class != TreeLab::Node
            raise "cannot attach node to itself" if node == self
            @root = node
            if(TreeLab::tree) 
                view_tree(TreeLab::tree)
            end
            return to_s
        end
        
        # checks whether the tree is empty
        def isEmpty?
            return root == nil
        end
        
        # counts number of nodes in the tree
        def noOfNodes
            if @root == nil
                return 0
            end
            return countNode(@root)
        end
        
        # counts the height of the tree
        def heightOfTree
            if (@root == nil)
                return 0
            else
                return countHeight(@root)
            end
        end
        
        # empties the tree
        def empty
            @root = nil
            if(TreeLab::tree) 
                view_tree(TreeLab::tree)
            end
            return true
        end
        
        def inspect
            return to_s
        end
        
        #    returns string representation of root node
        def to_s
            return nil if @root == nil
            return @root.to_s
        end
                    
        private
        #    private recursive method to count the number of nodes
        def countNode(node)
            if (node == nil)
                return 0
            end
            if (node.isLeaf?)
                return 1
            end
            return 1 + countNode(node.left) + countNode(node.right)
        end
        
        #    private recursive method to count the height of the tree
        def countHeight(node)
            if (node == nil)
                return 0
            end
            if (node.isLeaf?)
                return 1
            end
            return 1 + [countHeight(node.left),countHeight(node.right)].max
        end        
    end
    
    #    Node class used to represent each node in a binary tree
    class Node
    
        # left child of node
        attr_reader:left
        # right child of node
        attr_reader:right
        # value of node
        attr_accessor:value
        # attribute that represents the node in a binary tree canvas
        attr_accessor:circle
        
        def initialize(value)
            @value = value
        end 
        
        #    sets the left child node
        def setLeft(node)
            raise "Value must be a Node!" if node.class != TreeLab::Node
            raise "cannot attach node to itself" if node == self
            raise "left child is not empty" if @left != nil
            @left = node
            if(TreeLab::tree) 
                view_tree(TreeLab::tree)
            end
            return to_s
        end
        
        #    sets the right child node
        def setRight(node)
            raise "Value must be a Node!" if node.class != TreeLab::Node
            raise "cannot attach node to itself" if node == self
            raise "left child is not empty" if @right != nil
            @right = node
            if(TreeLab::tree) 
                view_tree(TreeLab::tree)
            end
            return to_s
        end
        
        #    deletes the left child node
        def delLeft
            @left = nil
            if(TreeLab::tree) 
                view_tree(TreeLab::tree)
            end
        end
        
        #    deletes the left child node
        def delRight
            @right = nil
            if(TreeLab::tree) 
                view_tree(TreeLab::tree)
            end
        end
        
        #    checks whether this node is a leaf node
        def isLeaf?
            return (left == nil && right == nil);
        end
        
        # return a string representation of the node
        def inspect
            return to_s
        end
        
        # returns a string representation of the node
        def to_s    
            s = "Value: "+value.to_s;
            if (left != nil)
                s += ", left child: "+left.value.to_s;
            else
                s += ", left child: nil"
            end
            if (right != nil)
                s += ", right child: "+right.value.to_s;
            else
                s += ", right child: nil"
            end
            return s;
        end
    end
    
    class BinarySearchTree < BinaryTree
        
        # :begin :search
        def search(key)
            if !stable
                p "Warning! Binary Search tree is not sorted correctly."
            end
            if(TreeLab::tree)
                @@visited = Array.new
                view_tree(self)
                bstsearch(@root, key)
            end
            return searchbst(@root, key)
        end
        # :end :search
        
        # :begin :insert
        def insert(key)
            if !stable
                p "Warning! Binary Search tree is not sorted correctly."
            end
            if(TreeLab::tree) 
                @@visited = Array.new
                view_tree(self)
                bstsearch(@root, key)
            end
            node = insertbst(@root, key)
            if(TreeLab::tree) 
                view_tree(TreeLab::tree)
            end
            
            return node
        end
        # :end :insert
        
        def stable
            @@stableChecker = Array.new
            inorder_check(@root)
            
            for i in 1..@@stableChecker.length-1
                return false if @@stableChecker[i] < @@stableChecker[i-1]
            end
            return true
        end
        
        private
        
        def inorder_check(root)
            if (root != nil) 
                inorder_check(root.left)
                @@stableChecker << root.value
                inorder_check(root.right)
            end
        end

        # :begin :insertbst
        def insertbst(root, key)
            if root.value == key
                return root
            elsif root.value > key
                if root.left != nil
                    return insertbst(root.left, key)
                else
                    node = Node.new(key)
                    root.setLeft(node)
                    return node
                end
            else
                if root.right != nil
                    return insertbst(root.right, key)
                else
                    node = Node.new(key)
                    root.setRight(node)
                    return node
                end
            end
        end
        # :end :insertbst
        
        # :begin :searchbst
        def searchbst(root, key)
            if root.value == key
                return root
            elsif root.left != nil and root.value > key
                return searchbst(root.left, key)
            elsif root.right != nil and root.value < key
                return searchbst(root.right, key)
            else
                return nil
            end
        end
        # :end :searchbst
    end
    
    private
    
    #private recursive method to draw the tree
    def preOrder(t, x, y, width)
        if (t != nil)
            t.circle = Canvas::Circle.new(x, y, 15)
            Canvas::Text.new(t.value, x, y, :anchor => :center)
            if (t.left != nil)
                Canvas::Line.new(x-15, y, x-(width/2), y+50-15)
            end
            preOrder(t.left, x-(width/2), y+50, width/2)
            if (t.right != nil)
                Canvas::Line.new(x+15, y, x+(width/2), y+50-15)
            end
            preOrder(t.right, x+(width/2), y+50, width/2)
        end
    end

    #    :begin :preorder_traverse
    def preorder_traverse(node)
        if (node != nil) 
            visit(node)
            preorder_traverse(node.left)
            preorder_traverse(node.right)
        end
    end
    #    :end :preorder_traverse
    
    #    :begin :postorder_traverse
    def postorder_traverse(node)
        if (node != nil) 
            postorder_traverse(node.left)
            postorder_traverse(node.right)
            visit(node)
        end
    end
    #    :end :postorder_traverse
    
    #    :begin :inorder_traverse
    def inorder_traverse(node)
        if (node != nil) 
            inorder_traverse(node.left)
            visit(node)
            inorder_traverse(node.right)
        end
    end
    #    :end :inorder_traverse
    
    def bstsearch(node, val)
        if (node != nil) 
            visit(node)
            if (val == node.value)
                
            elsif (val < node.value)
                bstsearch(node.left, val)
            else
                bstsearch(node.right, val)
            end
        end
    end
    
    #    :begin :visit
    def visit(node)
        @@visited << node.value
        Canvas::Text.new(@@visited.length, node.circle.coords[0]+11, node.circle.coords[1]+33, :font => 'bucketfont')
        node.circle.fill = 'red'
        sleep(@@sleep)
        node.circle.fill = 'green'
    end
    #    :end :visit
    
    def updateVisitedText(value)
        if (value == nil)
        end
    end
end
